/* minimization of L(#)-convex function
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"

static int nbit(unsigned x) {
    int i;

    for (i=0; x!=0; i++) x &= x-1;
    return i;
}

enum { TYPE_FIRST, TYPE_LAST, TYPE_MINUS, TYPE_PLUS, TYPE_RANDOM,
       TYPE_KOLMOGOROV };
int neighbor_type = TYPE_FIRST;

static int is_improved(int min_set, int signed_set, int sign) {
    switch (neighbor_type) {
    case TYPE_FIRST: return FALSE;
    case TYPE_LAST:  return TRUE;
    case TYPE_MINUS: return (nbit(min_set)*sign > nbit(signed_set)*sign);
    case TYPE_PLUS:  return (nbit(min_set)*sign < nbit(signed_set)*sign);
    case TYPE_RANDOM:
    case TYPE_KOLMOGOROV: return (irandom(0,1) == 1);
    default: assert(0!=0);
    }
}

/* returns  0 if no improvment was found. */
static long best_improve2(int dim, double func(int,int*), int x[],
                          double *min_f, const int sign, const int fix_index) {
    const unsigned long loop = 1L << dim;
    const unsigned long fix_set = (fix_index>=0) ? (1L << fix_index):(0);
    int i;
    long j, set, min_set = 0;
    double f;

    for (set=1; set<loop; set++) {
        if ((set & fix_set) != 0) set += fix_set;
        for (i=0,j=1; j<=set; i++,j<<=1) if ((j&set) != 0) x[i] += sign;
        f = func(dim, x);  oracle_call++;
        if (f < *min_f ||
            (f == *min_f && is_improved(min_set, set, sign))) {
            *min_f = f;  min_set = set;  improve_call++;
        }
        for (i=0,j=1; j<=set; i++,j<<=1) if ((j&set) != 0) x[i] -= sign;
    }
    return min_set;
}

/* returns  0 if no improvment was found. */
/* returns >0 if improvment was found in plus set */
/* returns <0 if improvment was found in minus set */
static long best_improve(int dim, double func(int,int*), int x[],
                         const int fix_index) {
    long set_plus, set_minus;
    double f, f_plus, f_minus;

    iteration++;
    f = f_plus = f_minus = func(dim, x);
    set_plus  =  best_improve2(dim, func, x, &f_plus,  +1, fix_index);
    set_minus = -best_improve2(dim, func, x, &f_minus, -1, fix_index);

    if (f_plus == f && f_minus == f) return 0;

    if (neighbor_type == TYPE_KOLMOGOROV) {
        if (f_plus  == f) return set_minus;
        if (f_minus == f) return set_plus;
        return (irandom(0,1)==0 ? set_plus : set_minus);
    }

    if (f_plus < f_minus) return set_plus;
    if (f_plus > f_minus) return set_minus;
    switch (neighbor_type) {
    case TYPE_FIRST: case TYPE_MINUS:
        return set_plus;
    case TYPE_LAST:  case TYPE_PLUS:
        return set_minus;
    case TYPE_RANDOM:
        return (irandom(0,1)==0 ? set_plus : set_minus);
    default: assert(0!=0);
    }
}


static void steepest_descent(int dim, double func(int,int*), int x[],
			     const int fix_index) {
    int i, sign;
    long j, set;

    while ((set=best_improve(dim, func, x, fix_index)) != 0) {
        if (set > 0) sign = +1;
        else {       sign = -1;  set = -set; }
        for (i=0,j=1; j<=set; i++,j<<=1) if ((j&set) != 0) x[i]+=sign;
#ifdef DEBUG
        status_print(dim, func, x);
#endif
    }
}

/* returns the minimum value of func(x)    */
/* INPUT  func   : pointer to the function */
/* INPUT  init[] : the initial solution    */
/* OUTPUT init[] : the minimized solution  */
double lconv_minimize_type(int dim, double func(int,int*), int init[],
			   const int fix_index, const int type) {
    neighbor_type = type;
    counter_reset();
    if (!is_feasible(dim, func, init)) return PLUS_INFINITY;
    status_print(dim, func, init);
    assert(LONG_MAX > pow(2, dim));

    steepest_descent(dim, func, init, fix_index);

    status_print(dim, func, init);
    counter_print();
    return func(dim, init);
}

double lgconv_minimize_type(int dim, double func(int,int*), int init[],
			    const int type) {
    return lconv_minimize_type(dim, func, init, -1, type);
}
